/******** Author: Juan Reyes L.        *****
********* juan.reyes.lopez@gmail.com   *****
********* Based on Rob Hess pf tracker *****
********* November 2011 UTFSM Chile    ****/

#include <opencv/cv.h>
#include "opencv2/highgui.hpp"
#include "opencv2/video/tracking.hpp"
#include "opencv2/imgproc.hpp"
#include "opencv2/videoio.hpp"
#include "opencv2/highgui.hpp"
#include <string.h>
#include <iostream>
#include "tracker.h"
#include "particle.h"

using namespace cv;
using namespace std;

int main(int argc, const char *argv[])
{
	VideoCapture cam;
	int partn = 100;
	bool videoFile=false;
	bool showParticles=false;
	float sd = SD;
	float scale_sd = SCALE_SD;
	cout << "Usage:\n" << argv[0] << " [-p particlesNumber] [-v videoPath] [-sp] [-sd particle_pos_sd] [-sds scale_factor_sd] [-h]\n";
	cout << "The -sp option means show particles" << endl; 	
	cout << "The -h option prints this instructions" << endl;
	if(argc > 1) {
		for(int j=1; j < argc; j++) {
			if(strcmp(argv[j],"-v") == 0) {
				cam = *new VideoCapture(argv[j+1]);
				if( !cam.isOpened() ) {
					cout << "Failed to open file " << argv[j+1] << endl;
					return 0;
				}
				videoFile=true;
			} else if(strcmp(argv[j],"-p") == 0) {
				partn = atoi(argv[j+1]);
			} else if(strcmp(argv[j],"-sd") == 0) {
				sd = atof(argv[j+1]);
			} else if(strcmp(argv[j],"-sds") == 0) {
				scale_sd = atof(argv[j+1]);
			} else if(strcmp(argv[j],"-sp") == 0) {
				showParticles = true;
			} else if(strcmp(argv[j],"-h") == 0) {
				return 0;
			}
		}	
	}
	if(!videoFile) {
		cam = *new VideoCapture(0);
		if( !cam.isOpened() ) {
			cout << "Failed to open camera" << endl;
			return 0;
		}

	}
	Mat img;
	Mat hsvImg;
	MatND hist1;
	char key = 10;
	cam >> img;
	namedWindow("pftracker",0);
	cvSetMouseCallback("pftracker",mouseEventHandler,&img);
	particle* pArr;
	particle best;
	bool pInit=false;
	
	while(key != 27) {
		cam >> img;
		key = cvWaitKey(33);
	
		//if the user select the target object	
		if(areaDefined) {
			cvtColor(img,hsvImg,CV_BGR2HSV);
			//executed just one time, to initalize
			if(!pInit) {
				hist1 = getHistogramHSV(hsvImg(*sArea));
				pArr = init_particles(sArea,&hist1,1,partn);
				pInit = true;
			}
			//update using gaussian random number generator	
			updateParticles(pArr,partn,hsvImg,hist1,sd,scale_sd);
			//select the best particle
			best = getBest(pArr,partn);
			//make copies of best particles and erase worsts
			pArr = resampleParticles(pArr,partn);			
			rectangle(img,
						Point(best.x-best.width/2*best.scale,best.y-best.height/2*best.scale),
					 	Point(best.x + best.width/2*best.scale,best.y + best.height/2*best.scale),
						Scalar(0,0,255) );

			//show all particles
			for(int p=0; p < partn && showParticles; p++) {
				best = pArr[p];
				rectangle(img,
						Point(best.x-best.width/2*best.scale,best.y-best.height/2*best.scale),
					 	Point(best.x + best.width/2*best.scale,best.y + best.height/2*best.scale),
						Scalar(255,0,0) );
			} 	
		} else {
			cvWaitKey(300);
		}
		imshow("pftracker",img);
	}	
	return 0;
}
